import .step_cbv
open tm type star step

--meta def HintsStar := [``(srefl), ``(sstep), ``(star_trans), ``(star_appl), ``(star_trans), ``(star_appr)]
--meta def HintsStep := [``(appl), ``(appr), ``(plusl), ``(plusr), ``(splus),``(sbeta)]

/-- Logical relations and semantic typing. -/

def V : Π {n}, type → tm n → Prop
| n tint (const _) := true 
| n tint _ := false
| n (tarrow A B) (lam e) := ∀ v, V A v → ∃ v', (subst_tm (scons v var_tm) e) ≻* v' /\ V B v'
| n _ _ := false 

def E {n} (A : type) (e : tm n) := ∃ v, e ≻* v ∧ @V n A v.

def G {n m} (Γ : ctx n) (σ : Fin n → tm m) := ∀ x, @V m (Γ x) (σ x).

def sem_types {n} (Γ : ctx n) (x : tm n) (A : type) := 
  ∀ {m} (σ: Fin n → tm m), G Γ σ → E A (subst_tm σ x).

notation Γ ` ⊨ `:50 x ` : `:50 A := sem_types Γ x A

/-- Compatibility and soundness. -/

lemma val_inclusion {n} (e : tm n) (A) : V A e → E A e 
:= assume h : V A e, show  E A e, from exists.intro e ⟨ (srefl step), h ⟩ 

lemma val {n} {v : tm n} (A) : V A v → value v 
  := assume h : _, begin 
    cases A; cases v; unfold V at h,
    any_goals {contradiction <|> unfold value} 
end

lemma compat_var {n} (Γ : ctx n) (A) (x : Fin n) : 
  Γ ⊢ (var_tm x) : A → Γ ⊨ (var_tm x) : A := 
begin 
  intros h₁ m σ h₂, 
  apply val_inclusion, 
  arw, 
  cases h₁, aauto
end

lemma compat_lam {n} (Γ : ctx n) (A B) (e : tm (nat.succ n)) : 
  ((scons A Γ) ⊨ e : B) → Γ ⊨ (lam e) : (A ⤏ B) := 
begin 
  intros h₁ m σ h₂, 
  arw,
  apply val_inclusion, rw[V], 
  intros v hv, 
  arw,
  have hg : G (scons A Γ) (scons v σ), from 
  begin 
    intro x, cases x;  simp [scons]; aauto
  end,
  cases (h₁ _ _) with t, 
  existsi t, 
  all_goals { aauto }
end


lemma compat_app {n} (Γ : ctx n) (e₁) (e₂) (A B : type): 
  ( Γ ⊨ e₁ : A⤏B) → ( Γ ⊨ e₂ : A )→ Γ ⊨ app e₁ e₂ : B := 

assume h₁ h₂ _ σ q, show _, from 
begin 
  arw, --simp with substLemmas,
  unfold sem_types at *, 
  cases (h₁ _ q) with v₁ x, cases x with p₁ q₁, 
  cases (h₂ _ q) with v₂ x, cases x with p₂ q₂, 
  cases v₁; unfold V at q₁, 
  any_goals { cases q₁, },
  cases (q₁ v₂ q₂) with v₃ x, cases x with p₃ q₃,
  existsi v₃, split; aauto, 
  apply star_trans; aauto_w ``(star_appl), 
    { apply star_trans, 
      apply star_appr; aauto, 
      apply star_trans; aauto,
      apply star_sbeta, aauto_w ``(val), 
    },   
end 


   
lemma agree_plus {n} (e₁ e₂ : tm n) (n₁ n₂) : 
  e₁ ≻* const n₁ → e₂ ≻* const n₂ → plus e₁ e₂ ≻* const (n₁+n₂) :=
begin
  generalize h₁: (const n₁) = c₁,
  intros p, induction p, rw ← h₁, 

  generalize h₂ : (const n₂) = c₂,
  intros q, induction q, rw ← h₂, 

  { constructor, apply splus, apply srefl },
  { apply sstep; aauto_w ``(plusr), },
  { intro h, apply sstep; aauto_w ``(plusl), },
end

lemma compat_plus {n} (Γ : ctx n) (e₁ e₂) :
  (Γ ⊨ e₁ : tint) →  (Γ ⊨ e₂ : tint) → Γ ⊨ (plus e₁ e₂) : tint :=
begin 
  intros h₁ h₂ _ σ h₃, 
  arw, --rw subst_tm.equations._eqn_5,
  cases (h₁ _ h₃) with x q, cases q with h₄ h₅,
  cases (h₂ _ h₃) with y q, cases q with h₆ h₇, 
  cases x,  any_goals {unfold V at *, contradiction},  
  cases y,  any_goals {unfold V at *, contradiction},  
  apply (exists.intro (const (x_a+y_a))), 
  split, 
  { apply agree_plus; aauto, },
  simp [V]
end

lemma sem_soundness {n} (Γ : ctx n) (A e):
  Γ ⊢ e : A → Γ ⊨ e : A := 
begin 
  intros h, 
  induction h,
   { apply compat_var, constructor },
   { aauto_w ``(compat_app) },
   { aauto_w ``(compat_plus) },
   { apply compat_lam; aauto }, 
   { intros x σ h, 
     arw, --or: simp with substLemmas,
     existsi (const h_n), 
     split; aauto_w ``(srefl) 
   }
end

/-- Weak normalization. -/

lemma wn_cbv (e) (A) : 
  (empty_ctx ⊢ e : A) → ∃v, value v ∧ e ≻* v :=
assume h₁, have h₂ : _, from (@sem_soundness 0 _ _ _ h₁), 
  have h₃ :E A e, from
    begin 
      specialize (h₂ var_tm), 
      have h: E A (e[var_tm]) = E A e, from begin now_arw  end, 
      rw ← h, 
      apply h₂, 
      intros x, cases x
  end,
begin
  cases h₃ with x _, existsi x, 
  apply (and.intro), apply val, exact h₃_h.right, exact h₃_h.left, 
end 
