{-# LANGUAGE FlexibleContexts      #-}
{-# LANGUAGE MultiParamTypeClasses #-}
module Autosubst.Generator (genCode) where

import           Autosubst.Coq
import           Autosubst.GenM
import           Autosubst.Names
import           Autosubst.Syntax
import           Autosubst.Tactics
import           Autosubst.Types
import           Control.Monad.Except
import           Control.Monad.Reader
import           Control.Monad.State.Lazy
import           Data.List                as L

-- Generation of syntax from a signature
genVar :: TId -> SubstTy -> GenM InductiveCtors
genVar x n = do
  open_x <- isOpen x
  s <- finT_ x (sty_terms n)
  let t = [s] ==> sortType x n
  return [InductiveCtor (var_ x) [] (Just t) | open_x]

genCons :: TId -> SubstTy -> Constructor -> GenM InductiveCtor
genCons x n (Constructor cname pos) = do
  t <- do
    up_n_x <- mapM (\(Position bs y) -> liftM2 idApp (return y) (fmap sty_terms $ castUpSubst x (Position bs y) n)) pos
    return $  up_n_x ==> sortType x n
  return $ InductiveCtor (constructor_ cname) [] (Just t)

genBody :: TId -> GenM InductiveBody
genBody x = do
  cs <- constructors x
  (n,b) <- introScopeVar "n" x
  varCons <- genVar x n
  constructors <- mapM (genCons x n) cs
  return $ InductiveBody x (explicit_ b) (TermSort Type) (varCons ++ constructors)

genBodies :: [TId] -> GenM Inductive
genBodies xs = fmap Inductive (mapM genBody xs)

-- Generation of congruences.
genCongruence :: TId -> Constructor -> GenM Lemma
genCongruence x (Constructor cname pos) = do
    (m, bm) <- introScopeVar "m" x
    let s = getPattern "s" pos
    let t =  getPattern "t" pos
    let bs s = mapM (\(y, Position binders arg) -> do
                                                  xs' <-  castUpSubst x (Position binders arg) m
                                                  return $ BinderImplicitNameType [y] (sortType arg xs')) (zip s pos)
    bes <- bs s
    bt <- bs t
    let eqs = zipWith TermEq (map TermId s) (map TermId t)
    let eq = TermEq (idApp cname (sty_terms m ++ map TermId s)) (idApp cname (sty_terms m ++ map TermId t))
    return $ Lemma (congr_ cname) (bm ++ bes ++ bt) (eqs ==> eq) (ProofString "congruence.")

genCongruences :: TId -> GenM [Lemma]
genCongruences x = join $ fmap (mapM (genCongruence x)) (constructors x)

genUpRenS :: TId -> TId -> GenM Definition
genUpRenS y z = do
  (mn, bs) <- introRenScopeS ("m", "n")
  (xi, bxi) <- genRenS "xi" mn
  return $ Definition (upRen_ y z) (bs ++ bxi) Nothing (upRenT y z xi)

traversal :: TId -> SubstTy -> (TId -> String) ->  (Term -> Term) -> Binders -> [SubstTy] -> (Term -> Term)  -> (TId -> [Term] -> Term) -> GenM FixpointBody
traversal x scope name ret bargs args var_case sem = do
  let s = "s"
  cs <- constructors x
  open_x <- isOpen x
  let var_pattern = [Equation (PatCtor (var_ x) [s]) (var_case (TermId s)) | open_x]
  let cons_pattern (Constructor cname pos) = do
                                            let s = getPattern "s" pos
                                            args <- mapM (\pos -> mapM (castUpSubst x pos) args) pos
                                            return $ Equation (PatCtor cname s) (sem cname
                                                        (zipWith (\(arg,s) (Position bs p) -> idApp (name p) (map TermSubst arg ++ [TermId s])) (zip args s) pos))
  cons_pat <- mapM cons_pattern cs
  let t = TermMatch (MatchItem (TermId s) Nothing Nothing) (Just $ Todo "MBTerm") (var_pattern ++ cons_pat)
  return $ FixpointBody (name x) (bargs ++ [BinderNameType [s] (idApp x (sty_terms scope))]) (ret (TermId s)) t

genRenaming :: TId -> GenM FixpointBody
genRenaming x = do
  ((m,n),bs) <- introRenScope ("m", "n") x
  (xi,bxi) <- genRen x "xi" (m,n)
  toVarT <- toVar x xi
  traversal x m ren_ (const TermUnderscore) (bs ++ bxi) [xi]
                       (\s -> TermApp (var x n) [TermApp toVarT [s]])
                       (\x s -> idApp x (sty_terms n ++ s))

genRenamings :: [TId] -> GenM Fixpoint
genRenamings xs = fmap Fixpoint (mapM genRenaming xs)

upSubstT :: TId -> TId -> SubstTy -> Term -> GenM Term
upSubstT y z m sigma = do
    pat <- patternSId z y
    m <- upSubst y [z] m
    let sigma' = sigma >>> idApp (ren_ z) pat
    return $ if y == z then TermApp cons_ [TermApp (var y m) [varZero_], sigma'] else sigma'

genUpS :: TId -> TId -> GenM Definition
genUpS y z = do
  ((m,n), bs) <- introSubstScopeS ("m", "n") z
  (sigma, b_sigma) <- genSubstS "sigma" (m,n) z
  sigma <- upSubstT y z n sigma
  return $ Definition (up_ y z) (bs ++ b_sigma) Nothing sigma

genSubstitution :: TId -> GenM FixpointBody
genSubstitution x = do
  ((m, n), bmn) <- introRenScope ("m", "n") x
  (sigma, bs) <- genSubst x "sigma" (m,n)
  toVarT <- toVar x sigma
  traversal x m subst_ (const TermUnderscore) (bmn ++ bs) [sigma]
                       (\s -> TermApp toVarT [s])
                       (\x s -> idApp x (sty_terms n ++ s))

genSubstitutions :: [TId] -> GenM Fixpoint
genSubstitutions xs = fmap Fixpoint (mapM genSubstitution xs)

genUpId :: TId -> TId -> GenM Definition
genUpId y z = do
  (m, bm) <- introScopeVar "m" z
  m_var <- toVar z m
  (sigma, b_sigma) <- genSubstS "sigma" (m_var, m) z
  (eq, b_eq) <- genEq z "Eq" sigma (var z m)
  n <- tfresh "n"
  m <- upSubst z [y] m
  let ret = equiv_ (idApp (up_ y z) [sigma]) (var z m)
  shift <- patternSId z y
  let t n = ap_ [idApp (ren_ z) shift, TermApp eq [n]]
  let u = if y == z then matchFin_ (TermId n) t eq_refl_ else t (TermId n)
  return $ Definition (upId_ y z) (bm ++ b_sigma ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genIdL :: TId -> GenM FixpointBody
genIdL x = do
  (m, bm) <- introScopeVar "m" x
  (sigma, bs) <- genSubst x "sigma" (m, m)
  xs <- substOf x
  eqs' <- mapM (\y -> liftM2 idApp (return $ var_ y) (fmap sty_terms (castSubst x y m))) xs
  (eqs, beqs) <- genEqs x "Eq" (sty_terms sigma) eqs' (\x y s -> return $ idApp (upId_ y x) [TermUnderscore, s]) -- TODO: Generate ID in a sensible way.
  toVarT <- toVar x eqs
  let ret s = TermEq (idApp (subst_ x) (sty_terms sigma ++ [s])) s
  traversal x m idSubst_ ret (bm ++ bs ++ beqs) [sigma, eqs]
                       (\s -> TermApp toVarT [s])
                       (idApp . congr_)

genCompRenRen :: TId -> GenM FixpointBody
genCompRenRen x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (xi, bxi) <- genRen x "xi" (m,k)
  (zeta,bzeta) <- genRen x "zeta" (k,l)
  (rho, brho) <- genRen x "rho" (m, l)
  (eqs, beqs) <- genEqs x "Eq" (zipWith (>>>) (sty_terms xi) (sty_terms zeta)) (sty_terms rho) (\x y s -> return$ if x == y then idApp up_ren_ren_ [TermId ("xi" ++ x), TermId ("zeta" ++ x), TermId ("rho" ++ x), s] else s)
  toVarT <- toVar x eqs
  let ret s = TermEq (idApp (ren_ x) (sty_terms zeta ++ [idApp (ren_ x) $ sty_terms xi  ++ [s]])) (idApp (ren_ x) (sty_terms rho ++ [s]))
  traversal x m compRenRen_ ret (bkl ++ bm ++ bxi ++ bzeta ++ brho ++ beqs) [xi, zeta, rho, eqs]
                  (\n -> ap_ [var x l, TermApp toVarT [n]])
                  (idApp . congr_)

genUpRenSubst :: TId -> TId -> GenM Definition
genUpRenSubst y z = do
  (k, bk) <- introScopeVarS "k"
  (l, bl) <- introScopeVarS "l"
  (m, bm) <- introScopeVar "m" z
  (xi, bxi) <- genRenS "xi" (k, l)
  (tau, btau) <- genSubstS "tau" (l, m) z
  (theta, btheta) <- genSubstS "theta" (k, m) z
  m_var <- toVar z m
  (eq, b_eq) <- genEq z "Eq" (xi >>> tau) theta
  n <- tfresh "n"
  m <- upSubst z [y] m
  let ret = equiv_ (idApp (upRen_ y z) [xi] >>> idApp (up_ y z) [tau] ) (idApp (up_ y z) [theta])
  shift <- patternSId z y
  let t n = ap_ [idApp (ren_ z) shift, TermApp eq [n]]
  let u = if y == z then matchFin_ (TermId n) t eq_refl_  else t (TermId n) -- TODO: Generalize the matchFIn thing.
  return $ Definition (up_ren_subst_ y z) (bk ++ bl ++ bm ++ bxi ++ btau ++ btheta ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genCompRenSubst :: TId -> GenM FixpointBody
genCompRenSubst x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (xi, bxi) <- genRen x "xi" (m,k)
  (zeta,bzeta) <- genSubst x "tau" (k,l)
  (rho, brho) <- genSubst x "theta" (m, l)
  (eqs, beqs) <- genEqs x "Eq" (zipWith (>>>) (sty_terms xi) (sty_terms zeta)) (sty_terms rho) (\x y s -> return $ idApp (up_ren_subst_ y x) [TermId ("xi" ++ x), TermId ("tau" ++ x), TermId ("theta" ++ x), s])
  toVarT <- toVar x eqs
  let ret s = TermEq (idApp (subst_ x) (sty_terms zeta ++ [idApp (ren_ x) $ sty_terms xi  ++ [s]])) (idApp (subst_ x) (sty_terms rho ++ [s]))
  traversal x m compRenSubst_ ret (bkl ++ bm ++ bxi ++ bzeta ++ brho ++ beqs) [xi, zeta, rho, eqs]
                  (\n -> TermApp toVarT [n])
                  (idApp . congr_)

genUpSubstRen :: TId -> TId -> GenM Definition
genUpSubstRen y z = do
  (k, bk) <- introScopeVarS "k"
  (l, bl) <- introScopeVar "l" z
  (m, bm) <- introScopeVar "m" z
  (sigma, bsigma) <- genSubstS "sigma" (k, l) z
  (zeta, bzeta) <- genRen z "zeta" (l, m)
  (theta, btheta) <- genSubstS "theta" (k, m) z
  (eq, b_eq) <- genEq z "Eq" (sigma >>> idApp (ren_ z) (sty_terms zeta)) theta
  n <- tfresh "n"
  m <- upSubst z [y] m
  zeta' <- upSubst z [y] zeta
  pat <- patternSId z y
  let ret = equiv_ (idApp (up_ y z) [sigma] >>> idApp (ren_ z) (sty_terms zeta')) (idApp (up_ y z) [theta])
  shift <- patternSId z y
  let t n = eqTrans_ (idApp (compRenRen_ z) (pat ++ sty_terms zeta' ++ zipWith (>>>) (sty_terms zeta) pat ++ map (const (TermAbs [BinderName "x"] eq_refl_)) pat ++ [ TermApp sigma [n]]))
                (eqTrans_ (eqSym_ (idApp (compRenRen_ z) (sty_terms zeta ++ pat ++ zipWith (>>>) (sty_terms zeta) pat ++ map (const ((TermAbs [BinderName "x"] eq_refl_))) pat ++ [ TermApp sigma [n]])))
                (ap_ [idApp (ren_ z) pat, TermApp eq [n]]))
  let u = if y == z then matchFin_ (TermId n) t eq_refl_ else t (TermId n)
  return $ Definition (up_subst_ren_ y z) (bk ++ bl ++ bm ++ bsigma ++ bzeta ++ btheta ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genCompSubstRen :: TId -> GenM FixpointBody
genCompSubstRen x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (sigma, bsigma) <- genSubst x "sigma" (m,k)
  (zeta,bzeta) <- genRen x "zeta" (k,l)
  (theta, btheta) <- genSubst x "theta" (m, l)
  xs <- substOf x
  sigmazeta <- mapM (\(y, sigma) -> do
                zeta' <- castSubst x y zeta
                return $ sigma >>> idApp (ren_ y) (sty_terms zeta')) (zip xs  (sty_terms sigma))
  (eqs, beqs) <- genEqs x "Eq" sigmazeta (sty_terms theta) (\z y s -> do
                                  sigma' <- toVar z sigma
                                  zeta' <- castSubst x z zeta
                                  theta' <- toVar z  theta
                                  return $ idApp (up_subst_ren_ y z) ([sigma'] ++ sty_terms zeta' ++[theta', s]))
  toVarT <- toVar x eqs
  let ret s = TermEq (idApp (ren_ x) (sty_terms zeta ++ [idApp (subst_ x) $ sty_terms sigma  ++ [s]])) (idApp (subst_ x) (sty_terms theta ++ [s]))
  traversal x m compSubstRen_ ret (bkl ++ bm ++ bsigma ++ bzeta ++ btheta ++ beqs) [sigma, zeta, theta, eqs]
                  (\n -> TermApp toVarT [n])
                  (idApp . congr_)

genUpSubstSubst :: TId -> TId -> GenM Definition
genUpSubstSubst y z = do
  (k, bk) <- introScopeVarS "k"
  (l, bl) <- introScopeVar "l" z
  (m, bm) <- introScopeVar "m" z
  (sigma, bsigma) <- genSubstS "sigma" (k, l) z
  (tau, btau) <- genSubst z "tau" (l, m)
  (theta, btheta) <- genSubstS "theta" (k, m) z
  (eq, b_eq) <- genEq z "Eq" (sigma >>> idApp (subst_ z) (sty_terms tau)) theta
  n <- tfresh "n"
  m <- upSubst z [y] m
  zeta' <- upSubst z [y] tau
  pat <- patternSId z y
  let ret = equiv_ (idApp (up_ y z) [sigma] >>> idApp (subst_ z) (sty_terms zeta')) (idApp (up_ y z) [theta])
  shift <- patternSId z y
  zs <- substOf z
  pat' <- mapM (\(tau, z') -> do
                      p' <- castSubst z z' (SubstScope pat)
                      return $ tau >>> (idApp (ren_ z') (sty_terms p')) ) (zip (sty_terms tau) zs)
  let t n = (eqTrans_ (idApp (compRenSubst_ z) (pat ++ sty_terms zeta'  ++ (map (\(z, p) -> p >>> z) (zip (sty_terms zeta') pat)) ++ map (const ((TermAbs [BinderName "x"] eq_refl_))) pat ++ [ TermApp sigma [n]]))
                (eqTrans_ (eqSym_ (idApp (compSubstRen_ z) (sty_terms tau ++ pat ++ pat' ++ map (const ((TermAbs [BinderName "x"] eq_refl_))) pat ++ [ TermApp sigma [n]])))
                (ap_ [(idApp (ren_ z) pat), TermApp eq [n]])))
  let u = if y == z then matchFin_ (TermId n) t eq_refl_ else t (TermId n)
  return $ Definition (up_subst_subst_ y z) (bk ++ bl ++ bm ++ bsigma ++ btau ++ btheta ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genCompSubstSubst :: TId -> GenM FixpointBody
genCompSubstSubst x = do
  ((k,l), bkl) <- introRenScope ("k", "l") x
  (m, bm) <- introScopeVar "m" x
  (sigma, bsigma) <- genSubst x "sigma" (m,k)
  (tau,btau) <- genSubst x "tau" (k,l)
  (theta, btheta) <- genSubst x "theta" (m, l)
  xs <- substOf x
  sigmatau <- mapM (\(y, sigma) -> do
                tau' <- castSubst x y tau
                return $ sigma >>> idApp (subst_ y) (sty_terms tau')) (zip xs  (sty_terms sigma))
  (eqs, beqs) <- genEqs x "Eq" sigmatau (sty_terms theta) (\z y s -> do
                                  sigma' <- toVar z sigma
                                  tau' <- castSubst x z tau
                                  theta' <- toVar z  theta
                                  return $ idApp (up_subst_subst_ y z) ([sigma'] ++ sty_terms tau' ++[theta', s]))
  toVarT <- toVar x eqs
  let ret s = TermEq (idApp (subst_ x) (sty_terms tau ++ [idApp (subst_ x) $ sty_terms sigma  ++ [s]])) (idApp (subst_ x) (sty_terms theta ++ [s]))
  traversal x m compSubstSubst_ ret (bkl ++ bm ++ bsigma ++ btau ++ btheta ++ beqs) [sigma, tau, theta, eqs]
                  (\n -> TermApp toVarT [n])
                  (idApp . congr_)

genUpExtRen :: TId -> TId -> GenM Definition
genUpExtRen y z = do
  (m, bm) <- introScopeVarS "m"
  (n, bn) <- introScopeVarS "n"
  (xi, bxi) <- genRenS "xi" (m, n)
  (zeta,bzeta) <- genRenS "zeta" (m,n)
  (eq, b_eq) <- genEq z "Eq" xi zeta
  let ret = equiv_ (idApp (upRen_ y z) [xi]) (idApp (upRen_ y z) [zeta])
  n <- tfresh "n"
  let t n = TermApp eq [n]
  let s = matchFin_ (TermId n) (\n -> ap_ [shift_, t n]) eq_refl_
  let u = if y == z then s else t (TermId n)
  return $ Definition (upExtRen_ y z) (bm ++ bn ++ bxi ++ bzeta ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genExtRen :: TId -> GenM FixpointBody
genExtRen x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (xi, bxi) <- genRen x "xi" (m,n)
  (zeta, bzeta) <- genRen x "zeta" (m,n)
  xs <- substOf x
  (eqs, beqs) <- genEqs x "Eq" (sty_terms xi) (sty_terms zeta) (\x y s -> return$  idApp (upExtRen_ y x) [TermId ("xi" ++ x), TermId ("zeta" ++ x), s]) -- TODO: Shouldn't this want SubsttTy instead of Terms?
  let ret s = TermEq (idApp (ren_ x) (sty_terms xi ++ [s])) (idApp (ren_ x) (sty_terms zeta ++ [s]))
  toVarT <- toVar x eqs
  traversal x m extRen_ ret (bmn ++ bxi ++ bzeta ++ beqs) [xi, zeta, eqs]
                  (\z -> ap_ [var x n, TermApp toVarT [z]])
                  (idApp . congr_)

genUpExt :: TId -> TId -> GenM Definition
genUpExt y z = do
  (m, bm) <- introScopeVarS "m"
  (n, bn) <- introScopeVar "n" z
  (sigma, bsigma) <- genSubstS "sigma" (m, n) z
  (tau,btau) <- genSubstS "tau" (m,n) z
  (eq, b_eq) <- genEq z "Eq" sigma tau
  let ret = equiv_ (idApp (up_ y z) [sigma]) (idApp (up_ y z) [tau])
  shift <- patternSId z y
  n <- tfresh "n"
  let t n = ap_ [idApp (ren_ z) shift, TermApp eq [n]]
  let u = if y == z then matchFin_ (TermId n) t eq_refl_ else t (TermId n)
  return $ Definition (upExt_ y z) (bm ++ bn ++ bsigma ++ btau ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genExt :: TId -> GenM FixpointBody
genExt x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (sigma, bsigma) <- genSubst x "sigma" (m,n)
  (tau, btau) <- genSubst x "tau" (m,n)
  xs <- substOf x
  (eqs, beqs) <- genEqs x "Eq" (sty_terms sigma) (sty_terms tau) (\x y s -> return$  idApp (upExt_ y x) [TermId ("sigma" ++ x), TermId ("tau" ++ x), s]) -- TODO: Shouldn't this want SubsttTy instead of Terms?
  let ret s = TermEq (idApp (subst_ x) (sty_terms sigma ++ [s])) (idApp (subst_ x) (sty_terms tau ++ [s]))
  toVarT <- toVar x eqs
  traversal x m ext_ ret (bmn ++ bsigma ++ btau ++ beqs) [sigma, tau, eqs]
                  (\z ->  TermApp toVarT [z])   -- TODO: I didn't need the renaming case for Up. Dive into that.
                  (idApp . congr_)

genUpRinstInst :: TId -> TId -> GenM Definition
genUpRinstInst y z = do
  (m, bm) <- introScopeVarS "m"
  (n, bn) <- introScopeVar "n" z
  n_var <- toVar z n
  (xi, bxi) <- genRenS "xi" (m, n_var)
  (sigma, bsigma) <- genSubstS "sigma" (m, n) z
  (eq, b_eq) <- genEq z "Eq" (xi >>> var z n) sigma
  n' <- upSubst z [y] n
  let ret = equiv_ (idApp (upRen_ y z) [xi] >>> var z n') (idApp (up_ y z) [sigma])
  shift <- patternSId z y
  let t n = ap_ [idApp (ren_ z) shift, TermApp eq [n]]
  n <- tfresh "n"
  let u = if y == z then matchFin_ (TermId n) t eq_refl_ else t (TermId n)
  return $ Definition (up_rinstInst_ y z) (bm ++ bn ++ bxi ++ bsigma ++ b_eq) (Just ret) (TermAbs [BinderName n] u)

genRinstInst :: TId -> GenM FixpointBody
genRinstInst x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (xi, bxi) <- genRen x "xi" (m,n)
  (sigma, bsigma) <- genSubst x "sigma" (m,n)
  xs <- substOf x
  xis' <- mapM (\(y, xi) -> do
    n' <- castSubst x y n
    return$  xi >>> var y n') (zip xs (sty_terms xi))
  (eqs, beqs) <- genEqs x "Eq" xis' (sty_terms sigma) (\x y s -> return $  idApp (up_rinstInst_ y x) [TermId ("xi" ++ x), TermId ("sigma" ++ x), s]) -- TODO: Make this last part general: Allow a function whihc takes SubstTy arguments and then automatically gives right terms.
  let ret s = TermEq (idApp (ren_ x) (sty_terms xi ++ [s])) (idApp (subst_ x) (sty_terms sigma ++ [s]))
  toVarT <- toVar x eqs
  traversal x m rinstInst_ ret (bmn ++ bxi ++ bsigma ++ beqs) [xi, sigma, eqs]
                  (\n -> TermApp toVarT [n])
                  (idApp . congr_)


-- TODO: Proof Term.
genLemmaInstId :: TId -> GenM Lemma
genLemmaInstId x = do
  (m, bm) <- introScopeVar "m" x
  let ret = TermEq (idApp (subst_ x) (sty_terms m)) id_
  return $ Lemma ("instId_" ++ x) bm ret (ProofString "Admitted")

genLemmaVarL :: TId -> GenM Lemma
genLemmaVarL x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (sigma,bsigma) <- genSubst x "sigma" (m,n)
  sigma' <- toVar x sigma
  let ret = TermEq ((idApp (subst_ x) (sty_terms sigma)) >>> (var x m)) sigma'
  return $ Lemma "name" (bmn ++ bsigma) ret (ProofString "Admitted")

genLemmaRenSubst :: TId -> GenM Lemma
genLemmaRenSubst x = do
  ((m,n), bmn) <- introRenScope ("m", "n") x
  (xi, bxi) <- genRen x "xi" (m,n)
  xs <- substOf x
  xis' <- mapM (\(y, xi) -> do
    n' <- castSubst x y n
    return$  xi >>> var y n') (zip xs (sty_terms xi))
  let ret = TermEq (idApp (ren_ x) (sty_terms xi)) (idApp (subst_ x) xis')
  return (Lemma ("rinstInst_" ++ x) bmn ret (ProofString "Admitted"))


-- MISSING: All the inferred lemmas.
genCodeT :: [TId] -> [(TId,TId)] -> GenM Sentences
genCodeT xs upList = do
  varSorts <- filterM isOpen xs -- Sorts which have variables
  substSorts <- filterM (const $ return True) xs -- FEATURE: Only sorts which require renaming / substitution.
  types <- genBodies xs
  congruences <- mapM genCongruences xs
  upRen <- mapM (uncurry genUpRenS) upList
  renamings <- genRenamings substSorts
  ups <- mapM (uncurry genUpS) upList
  substitutions <- genSubstitutions substSorts
  upId <- mapM (uncurry genUpId) upList
  idlemmas <- mapM genIdL substSorts
  extUpRen <- mapM (uncurry genUpExtRen) upList
  extRen <- mapM genExtRen substSorts
  extUp <- mapM (uncurry genUpExt) upList
  ext <- mapM genExt substSorts
  compRenRen <- mapM genCompRenRen substSorts
  upRenSubst <- mapM (uncurry genUpRenSubst) upList
  compRenSubst <- mapM genCompRenSubst substSorts
  upSubstRen <- mapM (uncurry genUpSubstRen) upList
  compSubstRen <- mapM genCompSubstRen substSorts
  upSubstSubst <- mapM (uncurry genUpSubstSubst) upList
  compSubstSubst <- mapM genCompSubstSubst substSorts
  upRinstInst <- mapM (uncurry genUpRinstInst) upList
  rinstInst <- mapM genRinstInst substSorts
  return $ [SentenceInductive types] ++ map SentenceLemma (concat congruences) -- FEATURE: Different intermediate values. --}
          ++ map SentenceDefinition upRen ++ [SentenceFixpoint renamings]
          ++ map SentenceDefinition ups ++ [SentenceFixpoint substitutions]
          ++ map SentenceDefinition upId ++ [SentenceFixpoint $ Fixpoint idlemmas]
          ++ map SentenceDefinition extUpRen ++ [SentenceFixpoint $ Fixpoint extRen]
          ++ map SentenceDefinition extUp ++ [SentenceFixpoint $ Fixpoint ext]
          ++ [SentenceFixpoint $ Fixpoint compRenRen]
          ++ map SentenceDefinition upRenSubst ++ [SentenceFixpoint $ Fixpoint compRenSubst]
          ++ map SentenceDefinition upSubstRen ++ [SentenceFixpoint $ Fixpoint compSubstRen]
          ++ map SentenceDefinition upSubstSubst ++ [SentenceFixpoint $ Fixpoint compSubstSubst]
          ++ map SentenceDefinition upRinstInst ++ [SentenceFixpoint $ Fixpoint rinstInst]

genCode :: [[TId]] -> GenM Sentences
genCode xs = do
  let all = [(x,y) | x <- concat xs, y <- concat xs]
  (_, code) <- foldM (\(ups,sentences) x -> do
                                  xs <- substOf (L.head x)
                                  let up_x = [(x, y) | x <- xs, y <- xs]
                                  code_x <- genCodeT x (L.filter (`elem` ups) up_x)
                                  return (L.filter (`notElem` up_x) ups, sentences ++ [code_x]))
                                  (all, []) xs
  return $ concat code
