{-# LANGUAGE FlexibleContexts      #-}
{-# LANGUAGE FlexibleInstances     #-}
{-# LANGUAGE MultiParamTypeClasses #-}
{-# LANGUAGE TypeSynonymInstances  #-}
module Autosubst.Lean where

-- Pretty Printing of Syntax to Lean
import           Autosubst.Syntax
import           Text.PrettyPrint.Leijen

-- Primitives
eqSym_ :: Term -> Term
eqSym_ s = idApp "eq.symm"  [s]

eqTrans_ :: Term -> Term -> Term
eqTrans_ s t = idApp "eq.trans"  [s,t]

-- base types, names
nat :: Term
nat = TermId "nat"

expr :: Term
expr = TermId "expr"

tac :: Term
tac = TermId "tactic"

bool :: Term
bool = TermId "bool"

tt :: Expr
tt = ExprId "tt"

ff :: Expr
ff = ExprId "ff"

none_ :: String
none_ = "none"

some_ :: String
some_ = "some"

up_ren_ :: Term
up_ren_ = TermId "up_ren"

succ_ :: Term -> Term
succ_ n = TermApp (TermId "nat.succ") [n]

fin_ :: Term -> Term
fin_ n = TermApp (TermId "Fin") [n]

-- forward composition
(>>>) :: Term -> Term -> Term
s >>> t = idApp "comp" [t, s]

eq_refl_ :: Term
eq_refl_ = TermId "by refl"

shift_ :: Term
shift_ = TermId "shift"

id_ :: Term
id_ = TermId "id"

cons_ :: Term
cons_ = TermId "scons"

varZero_ :: Term
varZero_ = TermId "var_zero"

matchFin_ :: Term -> (Term -> Term) -> Term -> Term
matchFin_ s f b = TermMatch (MatchItem s Nothing Nothing) Nothing [Equation (PatCtor some_ ["n"]) (f (TermId "n")), Equation (PatQualId none_) b]

ap_ :: Terms -> Term
ap_ = idApp "ap"

-- simplifier attribute: all definitions and lemmas are tagged with it
simp_ :: Doc
simp_ = text "@[substLemmas]"

-- LeanShow Class
class LeanShow a where
  leanShow :: a -> Doc

instance LeanShow String where
  leanShow = text

cTupled ::  [Doc] -> Doc -> Doc
cTupled [] sep  = empty
cTupled [x] sep = x
cTupled xs sep = encloseSep lparen rparen (sep <+> empty) xs

-- cons/comp definitions in fintype,
-- support for well-founded structural recursion in wftactic,
-- simplifier attribute
leanPreamble :: Doc
leanPreamble = text "import .fintype" <> line <> text "import .wftactic" <> line
  <> text "run_cmd mk_simp_attr `substLemmas"

instance LeanShow CBinder where
  leanShow (BinderName s) = leanShow s
  leanShow (BinderNameType s t) = parens $ hsep (map leanShow s) <+> text ":" <+> leanShow t
  leanShow (BinderImplicitName s) = text "{" <+> leanShow s <+> text "}"
  leanShow (BinderImplicitNameType s t) = text "{" <+> hsep (map leanShow s) <+> text ":" <+> leanShow t <+> text "}"

instance LeanShow Definition where
  leanShow (Definition name bs mb s) = text "def" <+> leanShow name <+> hsep (map leanShow bs) <+> (text ":" <+> leanShow mb) <+> text ":="
                                        <$$> indent 2 (leanShow s)

instance LeanShow Equation where
  leanShow (Equation pat s) = text "|" <+> leanShow pat <+> text ":=" <+> leanShow s

instance LeanShow Equations where
    leanShow s = vcat (map leanShow s)

-- recursive definitions
instance LeanShow FixpointL where
  leanShow (FixpointL [b]) =
    text "def" <+> hcat (punctuate line (map (nest 2 . leanShow) [b]))
    -- <> line <> text "attribute [reducible]" <+> (\(FixpointLBody nm bs t e) -> text nm) b

  leanShow (FixpointL bs) =
    text "mutual def" <+> hcat (punctuate comma (map (\(FixpointLBody nm bs t e) -> text nm) bs))
    <> line <> text "with" <+> hcat (punctuate (line <> text " with ") (map (nest 2 . leanShow ) bs))
    <> line <> text "using_well_founded wf_single_arg" -- <> line <> text "attribute [reducible]"
    -- <+> hcat (punctuate (line <> text "attribute [reducible]" <+> empty) (map (\(FixpointLBody nm bs t e) -> text nm) bs))

instance LeanShow FixpointLBody where
  leanShow (FixpointLBody name bs ret pat) = leanShow name <+> colon <+> text "Π"
    <+> hsep (map leanShow bs) <> comma <+> leanShow ret <$$> indent 2 (leanShow pat)

-- meta definitions
instance LeanShow MetaDefinition where
  leanShow (MetaDefinition name [] ret eqs) =
    text "meta def" <+> leanShow name <+> colon  <+> leanShow ret
                                              <$$> hcat (punctuate line (map (indent 2 . (\s ->text "|" <+>s) . leanShow) eqs))
  leanShow (MetaDefinition name bs ret eqs) =
    text "meta def" <+> leanShow name <+> colon <+> hsep (map leanShow bs) </> comma <+> leanShow ret <+> text ":="
                                              <$$>  hcat (punctuate line (map (indent 2 . (\s ->text "|" <+>s) . leanShow) eqs))


instance LeanShow Lemma where
  leanShow (Lemma name bs mb s) = text "lemma" <+> leanShow name <+> hsep (map leanShow bs) <+> (text ":" <+> leanShow mb) <+> text ":="
                                        <$$> leanShow s
instance LeanShow Proof where
  leanShow (ProofExact tm) = leanShow tm
  leanShow (ProofString s) = text "begin" <+> leanShow s <+> text "end"
  leanShow (ProofAssrt nm ty tm ps) = text "have" <+> leanShow nm <+> colon <+> leanShow ty
   <> comma <+> text "from" <+> leanShow tm <> comma <+> hcat (punctuate comma $ map leanShow ps)

-- inductive definitions
instance LeanShow Inductive where
  leanShow (Inductive [b]) = text "inductive"
    <+> hcat (punctuate line (map (nest 2 . leanShow ) [b]))
    <$$> text "open" <+> foldr ((<+>) .  (\(InductiveBody nm bs s cs) -> text nm) ) empty [b] -- <> line <> text "attribute [reducible]" <+>  (\(InductiveBody nm bs s cs) -> text nm) b

  leanShow (Inductive bs) = text "mutual inductive" <+> hcat (punctuate comma (map  (\(InductiveBody nm bs s cs) -> text nm) bs))
    <+> line <> text "with" <+> hcat (punctuate (line <> text " with ") (map (nest 2 . leanShow ) bs)) <$$> text "open" <+> hsep (map  (\(InductiveBody nm bs s cs) -> text nm) bs)
    -- <> line <> text "attribute [reducible]" <+> hcat (punctuate (line <> text "attribute [reducible]" <+> empty) (map  (\(InductiveBody nm bs s cs) -> text nm) bs))

instance LeanShow InductiveBody where
  leanShow (InductiveBody name bs s cs) = hang 2 (leanShow name </> hsep (map leanShow bs) </> colon <+> leanShow s <+> text "") <$$> vsep (map leanShow cs)

instance LeanShow InductiveCtor where
  leanShow (InductiveCtor name bs mt) = text "|" </> text name <+> colon <+> text "Π" <+> hcat (punctuate colon (map leanShow bs)) <> comma <+> leanShow mt

instance LeanShow MatchItem where
  leanShow (MatchItem s mb mt) = leanShow s

-- option to leave arguments implicit
instance LeanShow a => LeanShow (Maybe a) where
  leanShow Nothing  = text "_"
  leanShow (Just t) = leanShow t

instance LeanShow Pattern where
  leanShow (PatCtor s ids) = parens (leanShow s <+> hsep (map leanShow ids))
  leanShow (PatCtorEx s p) = leanShow s </> hsep (map leanShow p)
  leanShow (PatQualId s)   = leanShow s
  leanShow (PatUnderscore)   = text "_"
  leanShow s               = text "Implement patterns."

instance LeanShow Sentence where
  leanShow (SentenceDefinition s) = simp_ <> line <> leanShow s
  leanShow (SentenceInductive s)  = simp_ <> line <> leanShow s
  leanShow (SentenceFixpointL s)   = simp_ <> line <> leanShow s
  leanShow (SentenceMetaDefinition s)   = leanShow s
  leanShow (SentenceLemma s)      = simp_ <> line <> leanShow s
  leanShow _                      = text "TODO: Other sentence forms."

instance LeanShow Sort where
  leanShow Prop = text "Prop"
  leanShow Set  = text "Set"
  leanShow Type = text "Type"


instance LeanShow SubstTy where
  leanShow xs = hsep (map (\x -> case x of
                                                          TermId s -> leanShow s
                                                          TermSubst s -> leanShow s
                                                          _ -> parens $ leanShow x) (sty_terms xs))

leanShowParens :: Term -> Doc
leanShowParens (TermId s)    = leanShow s
leanShowParens (TermSubst s) = leanShow s
leanShowParens s             = parens $ leanShow s
leanShowParens' :: Expr -> Doc
leanShowParens' (ExprId s)    = leanShow s
leanShowParens' (ExprTuple es)= cTupled (map leanShow es) comma
leanShowParens' s             = parens $ leanShow s

-- terms
instance LeanShow Term where
  leanShow (TermImp s) = text "@" <> leanShow s
  leanShow (Todo s)           = text ("Todo." ++ s)
  leanShow (TermApp s ts)     = leanShowParens s <+> hsep (map (\x -> case x of
                                                          TermId s -> leanShow s
                                                          TermSubst s -> leanShow s
                                                          TupleTy s -> leanShow x
                                                          _ -> parens $ leanShow x) ts)
  leanShow (TermId s)         = leanShow s
  leanShow (TermFunction s t) = leanShow s <+> text "->" <+> leanShow t
  leanShow (TermSort s)       = leanShow s
  leanShow TermUnderscore     = text "_"
  leanShow (TermForall bs t)  = text "forall" <+> hsep (map leanShow bs) <> text "," <+> leanShow t
  leanShow (TermAbs bs t)     = text "λ" <+> hsep (map leanShow bs) <+> text "," <+> leanShow t
  leanShow (TermLet s t u)    = text "let" <+> leanShow t <+> text ":=" <+> leanShow s <+> text "in" <+> leanShow u
  leanShow (Tuple s)          = cTupled (map leanShow s) comma
  leanShow (TupleTy s)        = cTupled(map leanShow s) (text " ×")
  leanShow (TermMatch m _ s)  = text "match" <+> leanShow m <+> text "with" <$$> vcat (map leanShow s) <$$> text "end"
  leanShow (TermAnd s)        = encloseSep empty empty (text " /\\") (map (parens . leanShow) s)
  leanShow (TermEq s t)       = leanShow s <+> text "=" <+> leanShow t
  leanShow (TermSubst s)      = leanShow s
  leanShow (TermProof p)      = leanShow p
  leanShow s                  = text "TODO"

-- meta expressions
instance LeanShow Expr where
  leanShow (ExprApp e es)     = leanShow e <+> hsep (map leanShow es)
  leanShow (ExprId s)         = leanShow s
  leanShow (ExprIte i thn e)  = text "if" <+> parens (leanShow i) <>line <> text "then" <+> leanShow thn <> line <> line <> text "else" <+> leanShow e
  leanShow (ExprRet e)        = text "return" <+> leanShow e
  leanShow (ExprDo es)      = text "do" <+> hsep (punctuate comma (map leanShow es))
  leanShow (ExprLet e1 e2) = leanShow e1 <+> text "←" <+> leanShow e2
  leanShow (ExprTuple es)     = cTupled (map leanShow es) comma
  leanShow (ExprEq e1 e2)     = leanShow e1 <+> text ":=" <+> leanShow e2
  leanShow (ExprBEq e1 e2)   = leanShow e1 <+> text "=" <+> leanShow e2
  leanShow (ExprBAnd e1 e2)   = leanShow e1 <+> text "&&" <+> leanShow e2
  leanShow (ExprQuote p)      = text "`(" <> leanShow p <> text ")"
  leanShow (ExprDQuote p)     = text "ex ``(" <> leanShow p <> text ")"
  leanShow (ExprUnderscore)   = text "_"
  leanShow (ExprQp p)         = text "%%" <> leanShow p
