{-# LANGUAGE FlexibleContexts      #-}
{-# LANGUAGE MultiParamTypeClasses #-}
module Autosubst.Syntax where

import           Autosubst.GenM
import           Autosubst.Types
import           Data.List            as L

import           Control.Monad.Except
import           Control.Monad.Reader
import           Control.Monad.RWS    hiding ((<>))

-- Contains syntax for functions, inductive datatypes...

type Identifier = String

-- Annotation --------------------------------------------------
data Annotation = Struct (Identifier)
                deriving ( Eq,Ord,Show)
-- Assertion ---------------------------------------------------
data Assertion = Assertion (AssertionKeyword) (Identifier) (Binders) (Term)
-- AssertionKeyword --------------------------------------------
data AssertionKeyword = AssTheorem
                      | AssLemma
                      | AssRemark
                      | AssFact
                      | AssCorollary
                      | AssProposition
                      | AssDefinition
                      | AssExample
-- CBinder ------------------------------------------------------
-- TODO: Add constructor for index binders
data CBinder = BinderName (String)
            | BinderNameType ([String]) (Term)
            | BinderImplicitName (String)
            | BinderImplicitNameType ([String]) (Term)
-- Binders -----------------------------------------------------
type Binders = [CBinder]
-- ClassDeclaration --------------------------------------------
data ClassDeclaration = ClassDeclaration (Identifier) (Binders) (MbSort) (MethodDeclarations)
-- ClassInstance -----------------------------------------------
data ClassInstance = ClassInstance (Identifier) (Binders) (Identifier) (Terms) (Methods)
-- ContextHyp --------------------------------------------------
data ContextHyp = ContextHyp (Identifier) (Pattern)
-- ContextHyps -------------------------------------------------
type ContextHyps = [ContextHyp]

-- Definition --------------------------------------------------
data Definition = Definition (String) (Binders) (MbTerm) (Term)
-- Equation ----------------------------------------------------
data Equation = Equation (Pattern) (Term)
-- Equations ---------------------------------------------------
type Equations = [Equation]
-- Fixpoint ----------------------------------------------------
data Fixpoint = Fixpoint (FixpointBodies)
-- FixpointBodies ----------------------------------------------
type FixpointBodies = [FixpointBody]
-- FixpointBody ------------------------------------------------
data FixpointBody = FixpointBody (String) (Binders) (Term) (Term)
-- Fixpoint (Lean) ---------------------------------------------
data FixpointL = FixpointL (FixpointLBodies)
-- FixpointBodies (Lean) ---------------------------------------
type FixpointLBodies = [FixpointLBody]
-- FixpointBody (Lean) -----------------------------------------
data FixpointLBody = FixpointLBody (String) (Binders) (Term) (Equations)
-- MetaDefinition (Lean) ---------------------------------------
data MetaDefinition = MetaDefinition (String) (Binders) (Term) (Exprs)
-- MetaExpression (Lean) ---------------------------------------

-- Hint --------------------------------------------------------
data Hint = HintResolve (Terms)
          | HintRewrite (Terms)
          | HintRewriteRightToLeft (Terms)
          | HintConstructors (Identifiers)
-- Identifiers -------------------------------------------------
type Identifiers = [(Identifier)]
-- Inductive ---------------------------------------------------
data Inductive = Inductive (InductiveBodies)
-- InductiveBodies ---------------------------------------------
type InductiveBodies = [InductiveBody]
-- InductiveBody -----------------------------------------------
data InductiveBody = InductiveBody (String) (Binders) (Term) (InductiveCtors)
-- InductiveCtor -----------------------------------------------
data InductiveCtor = InductiveCtor (String) (Binders) (MbTerm)
-- InductiveCtors ----------------------------------------------
type InductiveCtors = [InductiveCtor]
-- Lemma --
data Lemma = Lemma (String) (Binders) (Term) (Proof)

data Proof = ProofExact Term
          | ProofString String
          | ProofAssrt String Term Term [Proof]

-- MatchItem ---------------------------------------------------
data MatchItem = MatchItem (Term) (MbName) (MbTerm)
-- MbAnnotation ------------------------------------------------
type MbAnnotation = Maybe (Annotation)
-- MbName ------------------------------------------------------
type MbName = Maybe (Name)
-- MbPattern ---------------------------------------------------
type MbPattern = Maybe (Pattern)
-- MbSort ------------------------------------------------------
type MbSort = Maybe (Sort)
-- MbTerm ------------------------------------------------------
type MbTerm = Maybe (Term)
-- Method ------------------------------------------------------
data Method = Method (Identifier) (Identifiers) (Term)
-- MethodDeclaration -------------------------------------------
data MethodDeclaration = MethodDeclaration (Identifier) (Binders) (Term)
-- MethodDeclarations ------------------------------------------
type MethodDeclarations = [MethodDeclaration]
-- Methods -----------------------------------------------------
type Methods = [Method]
-- Modifier ----------------------------------------------------
data Modifier = ModNone
              | ModLocal
              | ModGlobal
              deriving ( Eq,Ord,Show)
-- Name --------------------------------------------------------
data Name = NameIdent (Identifier)
          | NameUnderscore
          deriving ( Eq,Ord,Show)
-- Names -------------------------------------------------------
type Names = [Name]
-- Pattern -----------------------------------------------------
data Pattern = PatCtor (String) (Identifiers)
             | PatCtorEx (String) (Patterns)
             | PatAtCtor (String) (Identifiers)
             | PatAtCtorEx (String) (Patterns)
             | PatQualId (String)
             | PatUnderscore
             deriving ( Eq,Ord,Show)
-- Patterns ----------------------------------------------------
type Patterns = [Pattern]
-- PossiblyBracketedName ---------------------------------------
data PossiblyBracketedName = BracketedName (Name)
                           | BracedName (Name)
                           | NormalName (Name)
                           deriving ( Eq,Ord,Show)
-- PossiblyBracketedNames --------------------------------------
type PossiblyBracketedNames = [PossiblyBracketedName]
-- QualId ------------------------------------------------------
data QualId = Ident (Identifier)
            deriving ( Eq,Ord,Show)
-- Root --------------------------------------------------------
data Root = Root (Sentences)
-- Scheme ------------------------------------------------------
data Scheme = Scheme (SchemeBodies)
            deriving ( Eq,Ord,Show)
-- SchemeBodies ------------------------------------------------
type SchemeBodies = [SchemeBody]
-- SchemeBody --------------------------------------------------
data SchemeBody = SchemeInduction (Identifier) (Identifier)
                deriving ( Eq,Ord,Show)
-- Sentence ----------------------------------------------------
data Sentence = SentenceDefinition (Definition)
              | SentenceInductive (Inductive)
              | SentenceFixpoint (Fixpoint)
              | SentenceFixpointL (FixpointL)
              | SentenceMetaDefinition (MetaDefinition)
              | SentenceSection (Identifier) (Sentences)
              | SentenceOpaque (Identifier)
              | SentenceHint (Modifier) (Hint) (Identifiers)
              | SentenceScheme (Scheme)
              | SentenceCombinedScheme (Identifier) (Identifiers)
              | SentenceBlankline
              | SentenceArguments (Modifier) (QualId) (PossiblyBracketedNames)
              | SentenceClassDecl (ClassDeclaration)
              | SentenceClassInst (ClassInstance)
              | SentenceVerbatim (String)
              | SentenceContext (Binders)
              | SentenceLemma (Lemma)
-- Sentences ---------------------------------------------------
type Sentences = [Sentence]
-- Sort --------------------------------------------------------
data Sort = Prop
          | Set
          | Type
          deriving ( Eq,Ord,Show)
-- Substitution Objects -------------
data SubstTy =
  SubstScope Terms
  | SubstRen Terms
  | SubstSubst Terms
  | SubstEq Terms (TId -> TId -> Term -> GenM Term)
  | SubstConst Terms

sty_terms :: SubstTy -> Terms
sty_terms (SubstScope xs) = xs
sty_terms (SubstRen xs)   = xs
sty_terms (SubstSubst xs) = xs
sty_terms (SubstEq xs f)  = xs
sty_terms (SubstConst xs) = xs

-- Term --------------------------------------------------------
data Term =  TermImp Term
          | TermApp (Term) (Terms)
          | TermNum (Int)
          | TermQualId (QualId)
          | TermId (String)
          | TermSort (Sort)
          | TermFunction (Term) (Term)
          | TermAbs (Binders) (Term)
          | TermForall (Binders) (Term)
          | TermAnd (Terms)
          | TermEq (Term) (Term)
          | TermLet (Term) (Term) (Term) -- Get a pattern??
          | TermUnderscore
          | TermMatch (MatchItem) (MbTerm) (Equations)
          | TupleTy (Terms)
          | Tuple (Terms)
          | Todo (String)
          | TermSubst SubstTy
          | TermArg Term String Term
          | TermProof Proof
-- Terms -------------------------------------------------------
type Terms = [Term]
-- Variable ----------------------------------------------------
data Variable = Variable (Identifier) (Term)

type TmScope = Terms

-- MetaExpression (Lean) ---------------------------------------
data Expr = ExprApp (Expr) (Exprs)
          | ExprId (String)
          | ExprIte (Expr) (Expr) (Expr)
          | ExprRet (Expr)
          | ExprDo (Exprs)
          | ExprLet (Expr) (Expr)
          | ExprTuple (Exprs)
          | ExprQuote (Expr)
          | ExprDQuote (Expr)
          | ExprEq (Expr) (Expr)
          | ExprBEq (Expr) (Expr)
          | ExprBAnd (Expr) (Expr)
          | ExprUnderscore
          | ExprQp (Expr)
type Exprs = [Expr]

idApp :: String -> Terms -> Term
idApp s t = TermApp (TermId s) t


eIdApp :: String -> Exprs -> Expr
eIdApp s t = ExprApp (ExprId s) t

-- Substitution related
sortType :: TId -> SubstTy -> Term
sortType x n = TermApp (TermId x) [TermSubst n]

getTerms' :: CBinder -> Terms
getTerms' (BinderName s)                = [TermId s]
getTerms' (BinderNameType xs s)         = map TermId xs
getTerms' (BinderImplicitName s)        = []
getTerms' (BinderImplicitNameType xs s) = []

-- TODO: Special binders for renamings / scope variables etc?
getTerms :: Binders -> Terms
getTerms = concatMap getTerms'

(==>) :: [Term] -> Term -> Term
s ==> t = foldr (\s t -> TermFunction s t) t s

getTypes' :: CBinder -> Terms
getTypes' (BinderName s)                = []
getTypes' (BinderNameType xs s)         = map (const s) xs
getTypes' (BinderImplicitName s)        = []
getTypes' (BinderImplicitNameType xs s) = map (const s) xs

getTypes :: Binders -> Terms
getTypes = concatMap getTypes'

getNms' :: String -> CBinder -> String
getNms' pst (BinderName s)                = s
getNms' pst (BinderNameType xs s)         = concatMap (\s -> s++pst) xs
getNms' pst (BinderImplicitName s)        = s
getNms' pst (BinderImplicitNameType xs s) = concatMap (\s -> s++pst) xs

getNms :: Binders -> [String]
getNms = map (getNms' " ")

implicitBinder :: CBinder -> Binders
implicitBinder (BinderName s) = []
implicitBinder (BinderNameType s t) = []
implicitBinder (BinderImplicitName s) = [BinderImplicitName s]
implicitBinder (BinderImplicitNameType s t) = [BinderImplicitNameType s t]
implicitBinders :: Binders -> Binders
implicitBinders l = concatMap implicitBinder l
explicitBinder :: CBinder -> Binders
explicitBinder (BinderName s) = [BinderName s]
explicitBinder (BinderNameType s t) = map (\x -> BinderNameType [x] t) s
explicitBinder (BinderImplicitName s) = []
explicitBinder (BinderImplicitNameType s t) = []
explicitBinders :: Binders -> Binders
explicitBinders l = concatMap explicitBinder l
